@@ -1,5 +1,5 @@
 # This file is part of audioread.
-# Copyright 2011, Adrian Sampson.
+# Copyright 2012, Adrian Sampson.
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -17,6 +17,8 @@ pipe.
 """
 import subprocess
 import re
+import threading
+import select
 
 class FFmpegError(Exception):
     pass
@@ -30,6 +32,31 @@ class UnsupportedError(FFmpegError):
 class NotInstalledError(FFmpegError):
     """Could not find the ffmpeg binary."""
 
+class ReadTimeoutError(FFmpegError):
+    """Reading from the ffmpeg command-line tool timed out."""
+
+class ReaderThread(threading.Thread):
+    """A thread that consumes data from a filehandle. This is used to ensure
+    that a buffer for an input stream never fills up.
+    """
+    # It may seem a little hacky, but this is the most straightforward &
+    # reliable way I can think of to do this. select() is sort of
+    # inefficient because it doesn't indicate how much is available to
+    # read -- so I end up reading character by character.
+    def __init__(self, fh, blocksize=1024):
+        super(ReaderThread, self).__init__()
+        self.fh = fh
+        self.blocksize = blocksize
+        self.daemon = True
+        self.data = []
+
+    def run(self):
+        while True:
+            data = self.fh.read(self.blocksize)
+            if not data:
+                break
+            self.data.append(data)
+
 class FFmpegAudioFile(object):
     """An audio file decoded by the ffmpeg command-line utility."""
     def __init__(self, filename):
@@ -40,11 +67,26 @@ class FFmpegAudioFile(object):
             )
         except OSError:
             raise NotInstalledError()
+
+        # Read relevant information from stderr.
         self._get_info()
 
-    def read_data(self, block_size=4096):
+        # Start a separate thread to read the rest of the data from
+        # stderr.
+        stderr_reader = ReaderThread(self.proc.stderr)
+        stderr_reader.start()
+
+    def read_data(self, block_size=4096, timeout=10.0):
         """Read blocks of raw PCM data from the file."""
+        # Read from stdout on this thread.
         while True:
+            # Wait for data to be available or a timeout.
+            rready, _, xready = select.select((self.proc.stdout,),
+                                              (), (self.proc.stdout,),
+                                              timeout)
+            if not rready and not xready:
+                raise ReadTimeoutError()
+
             data = self.proc.stdout.read(block_size)
             if not data:
                 break
@@ -77,33 +119,51 @@ class FFmpegAudioFile(object):
         """Given relevant data from the ffmpeg output, set audio
         parameter fields on this object.
         """
-        self.samplerate = int(re.search(r'(\d+) hz', s).group(1))
-
-        mode = re.search(r'hz, ([^,]+),', s).group(1)
-        if mode == 'stereo':
-            self.channels = 2
+        # Sample rate.
+        match = re.search(r'(\d+) hz', s)
+        if match:
+            self.samplerate = int(match.group(1))
         else:
-            match = re.match(r'(\d+) ', mode)
-            if match:
-                self.channels = int(match.group(1))
-            else:
-                self.channels = 1
+            self.samplerate = 0
 
-        durparts = re.search(
+        # Channel count.
+        match = re.search(r'hz, ([^,]+),', s)
+        if match:
+            mode = match.group(1)
+            if mode == 'stereo':
+                self.channels = 2
+            else:
+                match = re.match(r'(\d+) ', mode)
+                if match:
+                    self.channels = int(match.group(1))
+                else:
+                    self.channels = 1
+        else:
+            self.channels = 0
+
+        # Duration.
+        match = re.search(
             r'duration: (\d+):(\d+):(\d+).(\d)', s
-        ).groups()
-        durparts = map(int, durparts)
-        duration = durparts[0] * 60 * 60 + \
-                   durparts[1] * 60 + \
-                   durparts[2] + \
-                   float(durparts[3]) / 10
-        self.duration = duration
+        )
+        if match:
+            durparts = map(int, match.groups())
+            duration = durparts[0] * 60 * 60 + \
+                       durparts[1] * 60 + \
+                       durparts[2] + \
+                       float(durparts[3]) / 10
+            self.duration = duration
+        else:
+            # No duration found.
+            self.duration = 0
 
     def close(self):
         """Close the ffmpeg process used to perform the decoding."""
         if hasattr(self, 'proc') and self.proc.returncode is None:
             self.proc.terminate()
-            self.proc.communicate()
+            # Flush the stdout buffer (stderr already flushed).
+            stdout_reader = ReaderThread(self.proc.stdout)
+            stdout_reader.start()
+            self.proc.wait()
 
     def __del__(self):
         self.close()
@@ -118,3 +178,4 @@ class FFmpegAudioFile(object):
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.close()
         return False
+
